/*

by Luigi Auriemma

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hldec.h"
#include "hlenc.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define    close    closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
    #include <pthread.h>
#endif

#ifdef WIN32
    #define quick_thread(NAME, ARG) DWORD WINAPI NAME(ARG)
    #define thread_id   DWORD
#else
    #define quick_thread(NAME, ARG) void *NAME(ARG)
    #define thread_id   pthread_t
#endif

thread_id quick_threadx(void *func, void *data) {
    thread_id       tid;
#ifdef WIN32
    if(!CreateThread(NULL, 0, func, data, 0, &tid)) return(0);
#else
    pthread_attr_t  attr;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    if(pthread_create(&tid, &attr, func, data)) return(0);
#endif
    return(tid);
}



#define VER     "0.1.1a"
#define PORT    27015
#define BUFFSZ    2048
#define INFOS   "\xff\xff\xff\xffinfostring\n\0"
#define TIMEOUT    5    //5 seconds



int                 sd1,
                    sd2,
                    plen;
struct    sockaddr_in    peer1,
                    peer2;



quick_thread(ctos, void *null);
void showinfostring(u_char *buff, int size);
void timeout(int sock);
u_int resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    int            err;
    u_char        buff[BUFFSZ],
                *hlserver;
    u_short     hlport = PORT;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(2,0), &wsadata);
#endif


    setbuf(stdout, NULL);

    fputs("\n"
        "Half-Life single-proxy forwarder with encoding/decoding functions "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\nUsage: %s <HL_server> [HL_server_port(default %d)]\n"
            "\nNOTE: IT SUPPORTS ONLY ONE CLIENT!\n"
            "\n", argv[0], PORT);
        exit(1);
    }

    hlserver = argv[1];
    if(argc > 2) hlport = atoi(argv[2]);

        /* 1 = client */
        /* 2 = server */

    sd1 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd1 < 0) std_err();
    sd2 = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd2 < 0) std_err();


    peer1.sin_addr.s_addr = INADDR_ANY;         /* listen for client */
    peer1.sin_port        = htons(PORT);
    peer1.sin_family      = AF_INET;
    peer2.sin_addr.s_addr = resolv(hlserver);   /* connect to the HL server */
    peer2.sin_port        = htons(hlport);
    peer2.sin_family      = AF_INET;
    plen                  = sizeof(peer1);

       /* GET INFORMATIONS FROM THE SERVER */

    printf("\nChecking the remote Half-Life server %s:%hu\n",
        inet_ntoa(peer2.sin_addr), ntohs(peer2.sin_port));
    err = sendto(sd2, INFOS, sizeof(INFOS) - 1, 0, (struct sockaddr *)&peer2, plen);
    if(err < 0) std_err();
    timeout(sd2);
    err = recvfrom(sd2, buff, BUFFSZ, 0, (struct sockaddr *)&peer2, &plen);
    if(err < 0) std_err();
    buff[err] = 0;
    showinfostring(buff, err);

        /* BIND THE 27015 PORT */

    printf("\nBinding UDP port %u\n\n", PORT);
    err = bind(sd1, (struct sockaddr *)&peer1, plen);
    if(err < 0) std_err();

        /* CREATE THE THREAD THAT WILL CHECK THE DATA SENT BY THE CLIENT */

    if(!quick_threadx(ctos, NULL)) {
        fputs("\nError: Cannot create the thread\n", stdout);
        exit(1);
    }

       /* CHECK THE DATA SENT BY THE SERVER */

    fputs("Server -> Client (active)\n", stdout);

    while(1) {
        err = recvfrom(sd2, buff, BUFFSZ, 0, (struct sockaddr *)&peer2, &plen);
        if(err < 0) std_err();

        if(memcmp(buff + 1, "\xff\xff\xff", 3)) {
            hldec(buff, err);

                    /*--------------*/
                    /* DO SOMETHING */
                    /*--------------*/

            hlenc(buff, err);
        }

        err = sendto(sd1, buff, err, 0, (struct sockaddr *)&peer1, plen);
        if(err < 0) std_err();
    }

    return(0);
}



quick_thread(ctos, void *null) {
    int err;
    u_char  buff[BUFFSZ];

    fputs("Client -> Server (active)\n", stdout);

    while(1) {
        err = recvfrom(sd1, buff, BUFFSZ, 0, (struct sockaddr *)&peer1, &plen);
        if(err < 0) std_err();


        if(memcmp(buff + 1, "\xff\xff\xff", 3)) {
            hldec(buff, err);

            if(!memcmp(buff + 8, "\x03new\0", 5)) {
                printf("New client: %s:%hu\n",
                    inet_ntoa(peer1.sin_addr), ntohs(peer1.sin_port));
            }

                /*--------------*/
                /* DO SOMETHING */
                /*--------------*/

            hlenc(buff, err);
        }

        err = sendto(sd2, buff, err, 0, (struct sockaddr *)&peer2, plen);
        if(err < 0) std_err();
    }
    return(0);
}



void showinfostring(u_char *buff, int size) {
    int        nt = 1,
            len;
    u_char    *string;

    fputs("\n--------------------------------------------------\n", stdout);

    len = strlen(buff);
    if(len < size) buff += len + 1;

    while(1) {
        string = strchr(buff, '\\');
        if(!string) break;

        *string = 0;

        /* \n or \t */
        if(!nt) {
            printf("%s: ", buff);
            nt++;
        } else {
            printf("%s\n", buff);
            nt = 0;
        }
        buff = string + 1;
    }
    printf("%s\n\n", buff);
}



void timeout(int sock) {
    struct    timeval    tout;
    fd_set    fd_read;
    int    err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) {
        fputs("\nError: Socket timeout, no answers received\n", stdout);
        exit(1);
    }
}



u_int resolv(char *host) {
    struct hostent *hp;
    u_int host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE)     {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            return 0;
        } else host_ip = *(u_int *)hp->h_addr;
    }

    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif



